"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformPlugin = void 0;
const core_1 = require("@babel/core");
const helper_plugin_utils_1 = require("@babel/helper-plugin-utils");
const babel_preset_1 = require("@linaria/babel-preset");
const shaker_1 = require("@linaria/shaker");
const core_2 = require("@griffel/core");
const astify_1 = require("./utils/astify");
const evaluatePaths_1 = require("./utils/evaluatePaths");
const validateOptions_1 = require("./validateOptions");
function getDefinitionPathFromMakeStylesCallExpression(callExpression) {
    const argumentPaths = callExpression.get('arguments');
    const hasValidArguments = Array.isArray(argumentPaths) && argumentPaths.length === 1;
    if (!hasValidArguments) {
        throw new Error('makeStyles() function accepts only a single param');
    }
    const definitionsPath = argumentPaths[0];
    if (!definitionsPath.isObjectExpression()) {
        throw definitionsPath.buildCodeFrameError('makeStyles() function accepts only an object as a param');
    }
    return definitionsPath;
}
/**
 * Checks that passed callee imports makesStyles().
 */
function isMakeStylesCallee(path, modules) {
    if (path.isIdentifier()) {
        return Boolean(modules.find(module => path.referencesImport(module.moduleSource, module.importName)));
    }
    return false;
}
/**
 * Checks if import statement import makeStyles().
 */
function hasMakeStylesImport(path, modules) {
    return Boolean(modules.find(module => path.node.source.value === module.moduleSource));
}
/**
 * Checks that passed declarator imports makesStyles().
 *
 * @example react_make_styles_1 = require('@griffel/react')
 */
function isRequireDeclarator(path) {
    const initPath = path.get('init');
    if (!initPath.isCallExpression()) {
        return false;
    }
    if (initPath.get('callee').isIdentifier({ name: 'require' })) {
        const args = initPath.get('arguments');
        return Array.isArray(args) && args.length === 1 && args[0].isStringLiteral({ value: '@griffel/react' });
    }
    return false;
}
/**
 * Rules that are returned by `resolveStyles()` are not deduplicated.
 * It's critical to filter out duplicates for build-time transform to avoid duplicated rules in a bundle.
 */
function dedupeCSSRules(cssRules) {
    Object.keys(cssRules).forEach(styleBucketName => {
        cssRules[styleBucketName] = cssRules[styleBucketName].filter((rule, index, rules) => rules.indexOf(rule) === index);
    });
    return cssRules;
}
exports.transformPlugin = (0, helper_plugin_utils_1.declare)((api, options) => {
    api.assertVersion(7);
    const pluginOptions = Object.assign({ babelOptions: {}, modules: [{ moduleSource: '@griffel/react', importName: 'makeStyles' }], evaluationRules: [
            { action: shaker_1.default },
            {
                test: /[/\\]node_modules[/\\]/,
                action: 'ignore',
            },
        ] }, options);
    (0, validateOptions_1.validateOptions)(pluginOptions);
    return {
        name: '@griffel/babel-plugin-transform',
        pre() {
            this.definitionPaths = [];
            this.calleePaths = [];
        },
        visitor: {
            Program: {
                enter() {
                    // Invalidate cache for module evaluation to get fresh modules
                    babel_preset_1.Module.invalidate();
                },
                exit(path, state) {
                    if (!state.importDeclarationPath && !state.requireDeclarationPath) {
                        return;
                    }
                    if (state.definitionPaths) {
                        // Runs Babel AST processing or module evaluation for Node once for all arguments of makeStyles() calls once
                        (0, evaluatePaths_1.evaluatePaths)(path, state.file.opts.filename, state.definitionPaths, pluginOptions);
                        state.definitionPaths.forEach(definitionPath => {
                            const callExpressionPath = definitionPath.findParent(parentPath => parentPath.isCallExpression());
                            const evaluationResult = definitionPath.evaluate();
                            if (!evaluationResult.confident) {
                                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                                const deoptPath = evaluationResult.deopt;
                                throw (deoptPath || definitionPath).buildCodeFrameError('Evaluation of a code fragment failed, this is a bug, please report it');
                            }
                            const stylesBySlots = evaluationResult.value;
                            const [classnamesMapping, cssRules] = (0, core_2.resolveStyleRulesForSlots)(stylesBySlots);
                            // TODO: find a better way to replace arguments
                            callExpressionPath.node.arguments = [(0, astify_1.astify)(classnamesMapping), (0, astify_1.astify)(dedupeCSSRules(cssRules))];
                        });
                    }
                    if (state.importDeclarationPath) {
                        const specifiers = state.importDeclarationPath.get('specifiers');
                        const source = state.importDeclarationPath.get('source');
                        specifiers.forEach(specifier => {
                            if (specifier.isImportSpecifier()) {
                                // TODO: should use generated modifier to avoid collisions
                                const importedPath = specifier.get('imported');
                                const importIdentifierPath = pluginOptions.modules.find(module => {
                                    return (module.moduleSource === source.node.value &&
                                        // 👆 "moduleSource" should match "importDeclarationPath.source" to skip unrelated ".importName"
                                        importedPath.isIdentifier({ name: module.importName }));
                                });
                                if (importIdentifierPath) {
                                    specifier.replaceWith(core_1.types.identifier('__styles'));
                                }
                            }
                        });
                    }
                    if (state.calleePaths) {
                        state.calleePaths.forEach(calleePath => {
                            calleePath.replaceWith(core_1.types.identifier('__styles'));
                        });
                    }
                },
            },
            // eslint-disable-next-line @typescript-eslint/naming-convention
            ImportDeclaration(path, state) {
                if (hasMakeStylesImport(path, pluginOptions.modules)) {
                    state.importDeclarationPath = path;
                }
            },
            // eslint-disable-next-line @typescript-eslint/naming-convention
            VariableDeclarator(path, state) {
                if (isRequireDeclarator(path)) {
                    state.requireDeclarationPath = path;
                }
            },
            // eslint-disable-next-line @typescript-eslint/naming-convention
            CallExpression(path, state) {
                /**
                 * Handles case when `makeStyles()` is `CallExpression`.
                 *
                 * @example makeStyles({})
                 */
                if (!state.importDeclarationPath) {
                    return;
                }
                const calleePath = path.get('callee');
                if (!isMakeStylesCallee(calleePath, pluginOptions.modules)) {
                    return;
                }
                state.definitionPaths.push(getDefinitionPathFromMakeStylesCallExpression(path));
                state.calleePaths.push(calleePath);
            },
            // eslint-disable-next-line @typescript-eslint/naming-convention
            MemberExpression(expressionPath, state) {
                /**
                 * Handles case when `makeStyles()` is inside `MemberExpression`.
                 *
                 * @example module.makeStyles({})
                 */
                if (!state.requireDeclarationPath) {
                    return;
                }
                const objectPath = expressionPath.get('object');
                const propertyPath = expressionPath.get('property');
                const isMakeStylesCall = objectPath.isIdentifier({ name: state.requireDeclarationPath.node.id.name }) &&
                    propertyPath.isIdentifier({ name: 'makeStyles' });
                if (!isMakeStylesCall) {
                    return;
                }
                const parentPath = expressionPath.parentPath;
                if (!parentPath.isCallExpression()) {
                    return;
                }
                state.definitionPaths.push(getDefinitionPathFromMakeStylesCallExpression(parentPath));
                state.calleePaths.push(propertyPath);
            },
        },
    };
});
//# sourceMappingURL=transformPlugin.js.map